from excitingtools.exciting_dict_parsers.species_parser import parse_species_xml


def test_parse_species_xml():
    """ Test parsing of species.xml files.
    """
    assert isinstance(species_str, str), (
        "Expect species parser to handle strings of XML data, "
        "due to use of decorator"
    )

    species_dict = parse_species_xml(species_str)

    assert set(species_dict) == {'species', 'muffin_tin', 'atomic_states', 'basis'}, 'Top level species file keys'
    assert species_dict['species'] == {
        'chemicalSymbol': 'Zn', 'name': 'zinc', 'z': -30.0, 'mass': 119198.678
    }
    assert species_dict['muffin_tin'] == {
        'rmin': 1e-06, 'radius': 2.0, 'rinf': 21.8982, 'radialmeshPoints': 600.0
    }

    assert isinstance(species_dict['atomic_states'], list), 'Atomic states stored as a list'
    atomic_states = [{'n': 1, 'l': 0, 'kappa': 1, 'occ': 2.00000,
                      'core': True}, {'n': 2, 'l': 0, 'kappa': 1, 'occ': 2.00000, 'core': True},
                     {'n': 2, 'l': 1, 'kappa': 1, 'occ': 2.00000,
                      'core': True}, {'n': 2, 'l': 1, 'kappa': 2, 'occ': 4.00000, 'core': True},
                     {'n': 3, 'l': 0, 'kappa': 1, 'occ': 2.00000,
                      'core': False}, {'n': 3, 'l': 1, 'kappa': 1, 'occ': 2.00000, 'core': False},
                     {'n': 3, 'l': 1, 'kappa': 2, 'occ': 4.00000,
                      'core': False}, {'n': 3, 'l': 2, 'kappa': 2, 'occ': 4.00000, 'core': False},
                     {'n': 3, 'l': 2, 'kappa': 3, 'occ': 6.00000,
                      'core': False}, {'n': 4, 'l': 0, 'kappa': 1, 'occ': 2.00000, 'core': False}]
    assert species_dict['atomic_states'] == atomic_states

    basis = species_dict['basis']
    assert set(basis) == {'default', 'custom', 'lo'}, 'Keys for basis'

    assert basis['default'] == [{'type': 'lapw', 'trialEnergy': 0.1500, 'searchE': True}]

    # Custom apw, lapw or apw+lo
    assert basis['custom'] == [
        {'l': 0, 'type': 'lapw', 'trialEnergy': 1.35670550183736, 'searchE': False},
        {'l': 1, 'type': 'lapw', 'trialEnergy': -2.69952312512447,
         'searchE': False}, {'l': 2, 'type': 'lapw', 'trialEnergy': 0.00, 'searchE': False},
        {'l': 3, 'type': 'lapw', 'trialEnergy': 1.000,
         'searchE': False}, {'l': 4, 'type': 'lapw', 'trialEnergy': 1.000, 'searchE': False},
        {'l': 5, 'type': 'lapw', 'trialEnergy': 1.000, 'searchE': False}
    ]

    # All explicitly specified LOs
    los = [{'l': 0, 'wf': [{'matchingOrder': 0, 'searchE': False, 'trialEnergy': -4.37848525995355},
                           {'matchingOrder': 1, 'searchE': False, 'trialEnergy': -4.37848525995355}]},
           {'l': 0, 'wf': [{'matchingOrder': 0, 'searchE': False, 'trialEnergy': 1.35670550183736},
                           {'matchingOrder': 1, 'searchE': False, 'trialEnergy': 1.35670550183736}]},
           {'l': 0, 'wf': [{'matchingOrder': 0, 'searchE': False, 'trialEnergy': 1.35670550183736},
                           {'matchingOrder': 0, 'searchE': False, 'trialEnergy': -4.37848525995355}]},
           {'l': 0, 'wf': [{'matchingOrder': 1, 'searchE': False, 'trialEnergy': 1.35670550183736},
                           {'matchingOrder': 2, 'searchE': False, 'trialEnergy': 1.35670550183736}]},
           {'l': 1, 'wf': [{'matchingOrder': 0, 'searchE': False, 'trialEnergy': -2.69952312512447},
                           {'matchingOrder': 1, 'searchE': False, 'trialEnergy': -2.69952312512447}]},
           {'l': 1, 'wf': [{'matchingOrder': 1, 'searchE': False, 'trialEnergy': -2.69952312512447},
                           {'matchingOrder': 2, 'searchE': False, 'trialEnergy': -2.69952312512447}]},
           {'l': 2, 'wf': [{'matchingOrder': 0, 'searchE': False, 'trialEnergy': 0.0},
                           {'matchingOrder': 1, 'searchE': False, 'trialEnergy': 0.0}]},
           {'l': 2, 'wf': [{'matchingOrder': 1, 'searchE': False, 'trialEnergy': 0.0},
                           {'matchingOrder': 2, 'searchE': False, 'trialEnergy': 0.0}]},
           {'l': 3, 'wf': [{'matchingOrder': 0, 'searchE': False, 'trialEnergy': 1.0},
                           {'matchingOrder': 1, 'searchE': False, 'trialEnergy': 1.0}]},
           {'l': 3, 'wf': [{'matchingOrder': 1, 'searchE': False, 'trialEnergy': 1.0},
                           {'matchingOrder': 2, 'searchE': False, 'trialEnergy': 1.0}]},
           {'l': 4, 'wf': [{'matchingOrder': 0, 'searchE': False, 'trialEnergy': 1.0},
                           {'matchingOrder': 1, 'searchE': False, 'trialEnergy': 1.0}]},
           {'l': 4, 'wf': [{'matchingOrder': 1, 'searchE': False, 'trialEnergy': 1.0},
                           {'matchingOrder': 2, 'searchE': False, 'trialEnergy': 1.0}]},
           {'l': 5, 'wf': [{'matchingOrder': 0, 'searchE': False, 'trialEnergy': 1.0},
                           {'matchingOrder': 1, 'searchE': False, 'n': 5}]},
           {'l': 5, 'wf': [{'matchingOrder': 1, 'n': 5, 'searchE': False},
                           {'matchingOrder': 2, 'n': 6, 'searchE': False}]}]
    
    assert len(basis['lo']) == 14, "Number of explicitly-defined local orbitals"
    assert set(basis['lo'][0]) == {'l', 'wf'}, "Attributes defining a local orbital"
    assert basis['lo'] == los


def test_parse_species_xml_different_ordering():
    species_dict = parse_species_xml(species_str_diff_order)

    assert set(species_dict) == {'species', 'muffin_tin', 'atomic_states', 'basis'}, 'Top level species file keys'

    # References
    ref_species = {'chemicalSymbol': 'C', 'name': 'carbon', 'z': -6.0, 'mass': 21894.16673}
    ref_muffin_tin = {'rmin': 1e-05, 'radius': 1.45, 'rinf': 21.0932, 'radialmeshPoints': 250.0}
    ref_atomic_states = [{'n': 1, 'l': 0, 'kappa': 1, 'occ': 2.0, 'core': False},
                         {'n': 2, 'l': 0, 'kappa': 1, 'occ': 2.0, 'core': False},
                         {'n': 2, 'l': 1, 'kappa': 1, 'occ': 1.0, 'core': False},
                         {'n': 2, 'l': 1, 'kappa': 2, 'occ': 1.0, 'core': False}]
    ref_basis = {'default': [{'type': 'lapw', 'trialEnergy': 0.15, 'searchE': False}],
                 'custom': [{'l': 0, 'type': 'apw+lo', 'trialEnergy': 0.15, 'searchE': True},
                            {'l': 1, 'type': 'apw+lo', 'trialEnergy': 0.15, 'searchE': True}],
                 'lo': [{'l': 0, 'wfproj': False,
                         'wf': [{'matchingOrder': 0, 'trialEnergy': 0.15, 'searchE': True},
                                {'matchingOrder': 1, 'trialEnergy': 0.15, 'searchE': True},
                                {'matchingOrder': 0, 'trialEnergy': -9.8, 'searchE': False}]}]}
    
    assert species_dict['species'] == ref_species, "species data disagrees"
    assert species_dict['muffin_tin'] == ref_muffin_tin, "muffin tin data disagrees"
    assert species_dict['atomic_states'] == ref_atomic_states, "atomic state data disagrees"
    assert species_dict['basis']['default'] == ref_basis['default'], "basis default data disagrees"
    assert species_dict['basis']['custom'] == ref_basis['custom'], "basis custom data disagrees"
    assert species_dict['basis']['lo'] == ref_basis['lo'], "basis lo data disagrees"


species_str = """<?xml version="1.0" encoding="utf-8"?>
<spdb xsi:noNamespaceSchemaLocation="../../xml/species.xsd" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <sp chemicalSymbol="Zn" name="zinc" z="-30.0000" mass="119198.6780">
    <muffinTin rmin="0.100000E-05" radius="2.0000" rinf="21.8982" radialmeshPoints="600"/>
    <atomicState n="1" l="0" kappa="1" occ="2.00000" core="true"/>
    <atomicState n="2" l="0" kappa="1" occ="2.00000" core="true"/>
    <atomicState n="2" l="1" kappa="1" occ="2.00000" core="true"/>
    <atomicState n="2" l="1" kappa="2" occ="4.00000" core="true"/>
    <atomicState n="3" l="0" kappa="1" occ="2.00000" core="false"/>
    <atomicState n="3" l="1" kappa="1" occ="2.00000" core="false"/>
    <atomicState n="3" l="1" kappa="2" occ="4.00000" core="false"/>
    <atomicState n="3" l="2" kappa="2" occ="4.00000" core="false"/>
    <atomicState n="3" l="2" kappa="3" occ="6.00000" core="false"/>
    <atomicState n="4" l="0" kappa="1" occ="2.00000" core="false"/>
    <basis>
      <default type="lapw" trialEnergy="0.1500" searchE="true"/>

      <custom l="0" type="lapw" trialEnergy="1.35670550183736" searchE="false"/>
      <lo l="0">
        <wf matchingOrder="0" trialEnergy="-4.37848525995355" searchE="false"/>
	<wf matchingOrder="1" trialEnergy="-4.37848525995355" searchE="false"/>
      </lo>
      <lo l="0">
        <wf matchingOrder="0" trialEnergy="1.35670550183736" searchE="false"/>
        <wf matchingOrder="1" trialEnergy="1.35670550183736" searchE="false"/>
      </lo>
      <lo l="0">
        <wf matchingOrder="0" trialEnergy="1.35670550183736" searchE="false"/>
	<wf matchingOrder="0" trialEnergy="-4.37848525995355" searchE="false"/>
      </lo>
      <lo l="0">
        <wf matchingOrder="1" trialEnergy="1.35670550183736" searchE="false"/>
        <wf matchingOrder="2" trialEnergy="1.35670550183736" searchE="false"/>
      </lo>

      <custom l="1" type="lapw" trialEnergy="-2.69952312512447" searchE="false"/>
      <lo l="1">
        <wf matchingOrder="0" trialEnergy="-2.69952312512447" searchE="false"/>
        <wf matchingOrder="1" trialEnergy="-2.69952312512447" searchE="false"/>
      </lo>
      <lo l="1">
	<wf matchingOrder="1" trialEnergy="-2.69952312512447" searchE="false"/>
	<wf matchingOrder="2" trialEnergy="-2.69952312512447" searchE="false"/>
      </lo>

      <custom l="2" type="lapw" trialEnergy="0.00" searchE="false"/>
      <lo l="2">
	<wf matchingOrder="0" trialEnergy="0.00" searchE="false"/>
        <wf matchingOrder="1" trialEnergy="0.00" searchE="false"/>
      </lo>
      <lo l="2">
	<wf matchingOrder="1" trialEnergy="0.00" searchE="false"/>
        <wf matchingOrder="2" trialEnergy="0.00" searchE="false"/>
      </lo>

      <custom l="3" type="lapw" trialEnergy="1.000" searchE="false"/>
      <lo l="3">
	<wf matchingOrder="0" trialEnergy="1.000" searchE="false"/>
	<wf matchingOrder="1" trialEnergy="1.000" searchE="false"/>
      </lo>
      <lo l="3">
        <wf matchingOrder="1" trialEnergy="1.000" searchE="false"/>
        <wf matchingOrder="2" trialEnergy="1.000" searchE="false"/>
      </lo>

      <custom l="4" type="lapw" trialEnergy="1.000" searchE="false"/>
      <lo l="4">
	<wf matchingOrder="0" trialEnergy="1.000" searchE="false"/>
	<wf matchingOrder="1" trialEnergy="1.000" searchE="false"/>
      </lo>
      <lo l="4">
        <wf matchingOrder="1" trialEnergy="1.000" searchE="false"/>
        <wf matchingOrder="2" trialEnergy="1.000" searchE="false"/>
      </lo>

      <custom l="5" type="lapw" trialEnergy="1.000" searchE="false"/>
      <lo l="5">
	<wf matchingOrder="0" trialEnergy="1.000" searchE="false"/>
  <wf matchingOrder="1" n="5" searchE="false"/>      
      </lo>
      <lo l="5">
        <wf matchingOrder="1" n="5" searchE="false"/>
        <wf matchingOrder="2" n="6" searchE="false"/>
      </lo>

    </basis>
  </sp>
</spdb>
"""

species_str_diff_order = """<?xml version="1.0" encoding="UTF-8"?>
<spdb xsi:noNamespaceSchemaLocation="../../xml/species.xsd" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <sp chemicalSymbol="C" name="carbon" z="-6.00000" mass="21894.16673">
    <atomicState n="1" l="0" kappa="1" occ="2.00000" core="false"/>
    <atomicState n="2" l="0" kappa="1" occ="2.00000" core="false"/>
    <atomicState n="2" l="1" kappa="1" occ="1.00000" core="false"/>
    <atomicState n="2" l="1" kappa="2" occ="1.00000" core="false"/>
    <basis>
      <default type="lapw" trialEnergy="0.1500" searchE="false"/>
      <custom l="0" type="apw+lo" trialEnergy="0.1500" searchE="true"/>
      <lo l="0" wfproj="false">
	<wf matchingOrder="0" trialEnergy="0.1500" searchE="true"/>
	<wf matchingOrder="1" trialEnergy="0.1500" searchE="true"/>
	<wf matchingOrder="0" trialEnergy="-9.8" searchE="false"/>
      </lo>
      <custom l="1" type="apw+lo" trialEnergy="0.1500" searchE="true"/>
    </basis>
    <muffinTin rmin="0.100000E-04" radius="1.4500" rinf="21.0932" radialmeshPoints="250"/>
  </sp>
</spdb>
"""



"""Dynamic Class Construction

class_definitions defined as {'class name':
                                   {'bases': '(ParentClass1, ParentClass2, ...)',
                                    'attributes':
                                             {'attribute1', 'attribute1',
                                              'method1': 'method1'
                                             }
                                    }
                             }

where:
 * 'class name' defines the name of the class, when calling in Python.
 *  'bases' define any parent classes for "class name" to inherit from, given in a tuple.
 *  'attributes' defines any attributes (data or methods) to include in "class name".

ALL data should be defined as a string, as it will get rendered as a Python-valid string,
which subsequently gets interpreted at run-time.
"""
from typing import Dict, Union, List, Optional, Iterator

from excitingtools.exciting_dict_parsers.input_parser import special_tags_to_parse_map
from excitingtools.utils import valid_attributes
from excitingtools.utils.valid_attributes import input_valid_subtrees


def class_constructor_string(class_definitions: Dict[str, Union[dict, str]]) -> str:
    """ Given a dictionary, return a Python-interpretable string of one or more
    class definitions.

        Example function argument:
        class_definitions = {'class name':
                                  {'bases': '(ParentClass1, ParentClass2)',
                                   'attributes':
                                            {'attribute1', 'attribute1',
                                             'method1': 'method1'
                                            }
                                   }
                            }

    :param class_definitions: A dict containing the class name, attributes and parent classes,
    all in string form.
    :return: class_definition_str: A Python-interpretable string of class definitions using
    the type() constructor.
    """
    # first import the base class
    class_definition_str = 'from excitingtools.input.base_class import ExcitingXMLInput \n'
    for class_name, bases_and_attributes in class_definitions.items():
        # Parent class (or classes)
        bases = bases_and_attributes['bases']

        # Attributes and methods (including private attributes)
        attributes = bases_and_attributes['attributes']

        class_definition_str += \
            f"Exciting{class_name}Input = type('Exciting{class_name}Input', {bases}, {attributes}) \n"

    return class_definition_str


def give_class_dictionary(name: str) -> dict:
    """ Gives class dictionary with inheritance and further properties.

    :param name: name of class
    :return: dict with definition
    """
    return {"bases": '(ExcitingXMLInput, )',
            "attributes": {"__doc__": f"Class for exciting {name} input.",
                           "__module__": "excitingtools.input.input_classes",
                           'name': name}}


def class_name_uppercase(name: str) -> str:
    """ Converts the name in a string which better fits as class name, capitalizing the first letter and
    leaving the rest unchanged.
    Exceptions are groundstate -> GroundState, xs -> XS, gw -> GW, bandstructure -> BandStructure, eph -> EPH

    :param name: name of the class/tag in original notation
    :return: name usually capitilized, see exceptions
    """
    exceptions = {"groundstate": "GroundState", "xs": "XS", "gw": "GW", "bandstructure": "BandStructure", "eph": "EPH"}
    if name in exceptions:
        return exceptions[name]
    return name[0].upper() + name[1:]


def get_all_valid_subtrees(valid_xml_tags: Optional[List[str]]) -> Optional[Iterator[str]]:
    """ Get recursively all valid xml (sub-)trees (tags) from exciting.

    :param valid_xml_tags: valid xml tags
    :return: list of all valid xml tags of all trees and subtrees
    """
    if not valid_xml_tags:
        return
    for valid_tag in valid_xml_tags:
        yield valid_tag
        tag_valid_subtrees = valid_attributes.__dict__.get(f"{valid_tag}_valid_subtrees")
        yield from get_all_valid_subtrees(tag_valid_subtrees)


def generate_classes_str() -> str:
    """ Generate string to execute by python. For all standard input classes.

    :return: python string
    """
    non_standard_input_classes = set(special_tags_to_parse_map)
    all_valid_subtrees = set(get_all_valid_subtrees(input_valid_subtrees))
    standard_input_classes = all_valid_subtrees - non_standard_input_classes

    class_definitions = {class_name_uppercase(cls): give_class_dictionary(cls) for cls in standard_input_classes}
    return class_constructor_string(class_definitions)



"""
Tests for functions in lattice.py 
"""
import numpy as np

from excitingtools.structure.lattice import reciprocal_lattice_vectors, parallelepiped_volume, plane_transformation
from excitingtools.math.math_utils import triple_product


def test_parallelepiped_volume():

    # FCC lattice vectors with arbitrary lattice constant
    a = 3.2 * np.array([[0., 1., 1.], [1., 0., 1.], [1., 1., 0.]])

    triple_product_result = triple_product(a[:, 0], a[:, 1], a[:, 2])
    volume = parallelepiped_volume(np.array([a[:, 0], a[:, 1], a[:, 2]]))

    assert np.isclose(triple_product_result, 65.5360)
    assert np.isclose(volume, 65.5360)

    triple_product_result = triple_product(a[:, 0], a[:, 2], a[:, 1])
    volume = parallelepiped_volume(np.array([a[:, 0], a[:, 2], a[:, 1]]))

    assert np.isclose(triple_product_result, -65.5360), "triple product can be negative"
    assert np.isclose(volume, 65.5360), "volume is always defined as positve"


def test_reciprocal_lattice_vectors():
    r"""
    Test
        \mathbf{a}_i \cdot \mathbf{b}_j = 2 \pi \delta_{ij}

    """

    # fcc with arbitrary lattice constant
    a = 3.2 * np.array([[0., 1., 1.], [1., 0., 1.], [1., 1., 0.]])
    b = reciprocal_lattice_vectors(a)

    assert a.shape == (3, 3)
    assert b.shape == (3, 3)

    a_dot_b = np.transpose(a) @ b
    off_diagonal_indices = np.where(~np.eye(a_dot_b.shape[0], dtype=bool))
    off_diagonal_elements = a_dot_b[off_diagonal_indices]

    assert np.allclose(a_dot_b.diagonal(), 2 * np.pi)
    assert np.allclose(off_diagonal_elements, 0.)


def test_plane_transformation():
    """
    Test plane_transformation function.
    """
    rec_lat_vec = np.array([[-0.042506, -0.042506,  0.050235],
                         [-0.042506,  0.050235, -0.042506],
                         [-0.085013,  0.007728,  0.007728]]
                       )

    plot_vec = np.array([[-0.5881478,  0.5881478,  0.5881478],
                            [ 0.5881478, -0.5881478,  0.5881478],
                            [ 0.5881478,  0.5881478, -0.5881478]]
                          )
    # This is supposed to be 3x3 Identity
    transformation_matrix = plane_transformation(rec_lat_vec, plot_vec)

    assert np.allclose(transformation_matrix.dot(rec_lat_vec)[2,:], 0.0, atol=1e-6)



""" Input xml class. """
from pathlib import Path
from typing import Union

from excitingtools.input.base_class import ExcitingXMLInput
from excitingtools.input.input_classes import ExcitingGroundStateInput, ExcitingTitleInput
from excitingtools.input.structure import ExcitingStructure
from excitingtools.input.xml_utils import prettify_tag_attributes, xml_tree_to_pretty_str


class ExcitingInputXML(ExcitingXMLInput):
    """
    Container for a complete input xml file.
    """
    name = "input"
    _default_filename = "input.xml"
    structure: ExcitingStructure
    groundstate: ExcitingGroundStateInput
    title: ExcitingTitleInput

    def set_title(self, title: str):
        """ Set a new title. """
        self.__dict__["title"].title = title

    def to_xml_str(self) -> str:
        """Compose XML ElementTrees from exciting input classes to create an input xml string.

        :return: Input XML tree as a string, with pretty formatting.
        """
        return prettify_tag_attributes(xml_tree_to_pretty_str(self.to_xml()))

    def write(self, filename: Union[str, Path] = _default_filename):
        """Writes the xml string to file.

        :param filename: name of the file.
        """
        with open(filename, "w") as fid:
            fid.write(self.to_xml_str())
